#!/usr/bin/env python3

"""
Transfer completed BlobDirs and intermediate files.

Usage:
  blobtoolkit-pipeline transfer-completed --in PATH --out PATH [--bin PATH]

Options:
  --in PATH              Path to input directory.
  --out PATH             Path to output directory.
  --bin PATH             Path to bin directory. If specified, unwanted files will
                         be moved here rather than deleted.
"""

import glob
import gzip
import shutil
import tarfile
from pathlib import Path

import yaml
from docopt import DocoptExit
from docopt import docopt
from tolkein import tofetch
from tolkein import tofile
from tolkein import tolog

LOGGER = tolog.logger(__name__)


def compress_file(infile, outfile):
    """Compress a file using gzip."""
    LOGGER.info("Compressing %s to %s", infile, outfile)
    with gzip.open(outfile, "wb") as ofh:
        with open(infile, "rb") as ifh:
            shutil.copyfileobj(ifh, ofh)


def compress_content(indir, destdir):
    """Compress all files in directory."""
    Path(destdir).mkdir(parents=True, exist_ok=True)
    for infile in glob.glob(r"%s/*" % indir):
        if Path(infile).is_dir():
            outdir = "%s/%s" % (destdir, Path(infile).name)
            compress_content(infile, outdir)
        else:
            outfile = "%s/%s.gz" % (destdir, Path(infile).name)
            compress_file(infile, outfile)


def tar_directory(indir, destdir, *, compress=False):
    LOGGER.info("Archiving %s to %s.tar", indir, destdir)
    if compress:
        compress_content(indir, destdir)
    else:
        shutil.copytree(indir, destdir)
    with tarfile.open(destdir + ".tar", mode="w") as archive:
        archive.add(destdir, Path(destdir).name, recursive=True)
    shutil.rmtree(destdir)


def transfer_files(pattern, destdir, *, compress=False, rename=False):
    """Transfer files to output directory."""
    for file in glob.glob(pattern):
        filepath = Path(file)
        destfile = "%s/%s" % (destdir, filepath.name)
        if rename:
            destfile = destfile.replace(rename[0], rename[1])
        if filepath.is_dir():
            tar_directory(file, destfile, compress=compress)
            continue
        if compress:
            destfile += ".gz"
            compress_file(file, destfile)
            continue
        LOGGER.info("Moving %s to %s", file, destdir)
        shutil.move(file, destdir)


def create_static_directory(indir, staticdir):
    """Move image files to static directory."""
    Path(staticdir).mkdir(parents=True, exist_ok=True)
    for file in glob.glob(r"%s/*.png" % indir):
        shutil.move(file, staticdir)


def create_pipeline_directory(indir, outdir):
    """Move stats and log files to pipeline directory."""
    Path(f"{outdir}/logs").mkdir(parents=True, exist_ok=True)
    for file in glob.glob(r"%s/*.stats" % indir):
        shutil.move(file, outdir)
    for file in glob.glob(r"%s/*/logs/**/*.log" % indir):
        parts = file.split("/")
        outfile = "-".join([parts[1]] + parts[3:])
        shutil.move(file, f"{outdir}/logs/{outfile}")


def remove_unwanted_files(indir, bindir):
    """Move unwanted files to bin or delete."""
    LOGGER.info("Removing unwanted files")
    if bindir is not None:
        LOGGER.info("Moving %s to %s", indir, bindir)
        Path(bindir).mkdir(parents=True, exist_ok=True)
        shutil.move(indir, bindir)
    else:
        LOGGER.info("Deleting %s" % indir)
        shutil.rmtree(indir)


def main(rename=None):
    """Entry point."""
    docs = __doc__
    if rename is not None:
        docs = docs.replace("blobtoolkit-pipeline", rename)
    try:
        opts = docopt(docs)
    except DocoptExit as e:
        raise DocoptExit from e
    data = tofile.read_file("%s/config.yaml" % opts["--in"])
    meta = yaml.full_load(data)
    prefix = meta["assembly"]["prefix"]
    blobdir = prefix
    revision = meta.get("revision", 0)
    if revision:
        blobdir += ".%d" % int(revision)
    indir = opts["--in"]
    outdir = "%s/%s" % (opts["--out"], blobdir)
    bindir = opts.get("--bin", None)
    Path(outdir).mkdir(parents=True, exist_ok=True)
    create_pipeline_directory(indir, "%s/%s.pipeline" % (indir, prefix))
    transfer_files(r"%s/*.tar" % indir, outdir)
    transfer_files(r"%s/*.gz" % indir, outdir)
    transfer_files("%s/%s.pipeline" % (indir, prefix), outdir, compress=True)
    if Path("%s/view" % indir).is_dir():
        create_static_directory(
            "%s/view/%s" % (indir, blobdir), "%s/%s.static" % (indir, blobdir)
        )
        transfer_files("%s/view/%s" % (indir, blobdir), outdir, compress=True)
        transfer_files("%s/%s.static" % (indir, blobdir), outdir)
    else:
        transfer_files("%s/blobtools/%s" % (indir, blobdir), outdir, compress=True)
    transfer_files(r"%s/blastn/%s.*.out" % (indir, prefix), outdir, compress=True)
    transfer_files("%s/blobtools/%s.meta.yaml" % (indir, prefix), outdir, compress=True)
    transfer_files(r"%s/busco/%s.busco.*_odb10" % (indir, prefix), outdir)
    transfer_files("%s/config.yaml" % indir, outdir)
    transfer_files(
        "%s/cov_stats/%s.chunk_stats.tsv" % (indir, prefix), outdir, compress=True
    )
    transfer_files(
        r"%s/diamond/%s.*.out" % (indir, prefix),
        outdir,
        rename=("diamond.", "diamond_blastx."),
        compress=True,
    )
    transfer_files(
        "%s/diamond/%s.fasta.chunks" % (indir, prefix),
        outdir,
        rename=("fasta.chunks", "chunks.fasta"),
        compress=True,
    )
    transfer_files(
        r"%s/diamond_blastp/%s.*.out" % (indir, prefix),
        outdir,
        rename=("diamond.", "diamond_blastp."),
        compress=True,
    )
    transfer_files(r"%s/minimap/%s.*.bam" % (indir, prefix), outdir)
    transfer_files("%s/pipeline/%s" % (indir, prefix), outdir, compress=True)
    transfer_files(
        "%s/window_stats/%s.window_stats.tsv" % (indir, prefix), outdir, compress=True
    )
    transfer_files(
        "%s/windowmasker/%s.windowmasker.fasta" % (indir, prefix), outdir, compress=True
    )
    remove_unwanted_files(indir, bindir)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

"""
Track new data releases under a bioproject.

Usage:
  track_bioproject_accessions.py <BIOPROJECT> [--out /path/to/output/directory]

Options:
  --out PATH             Path to output directory [default: .]
"""

import os
from time import sleep

from docopt import docopt
from tolkein import tofetch
from tolkein import tofile
from tolkein import tolog

LOGGER = tolog.logger(__name__)

ENA_API = "https://www.ebi.ac.uk/ena/portal/api"


def fetch_bioproject_children(
    bioproject, *, projects=None, children=None, new_projects=None
):
    """
    Fetch children of a bioproject.
    """
    if projects is None:
        projects = []
    if new_projects is None:
        new_projects = {}
    if children is None:
        children = []
    url = "%s/search?result=study&query=parent_study%%3D%%22%s%%22&format=tsv" % (
        ENA_API,
        bioproject,
    )
    result = tofetch.fetch_url(url)
    if result and result is not None:
        for line in result.split("\n")[1:]:
            if line and "\t" in line:
                child_accession, description = line.split("\t")
                if child_accession in projects:
                    continue
                if "genome assembly" in description:
                    if "alternate haplotype" not in description:
                        children.append(child_accession)
                        new_projects.update({child_accession: bioproject})
                else:
                    sleep(0.5)
                    LOGGER.info(
                        "Fetching nested accessions under bioproject %s"
                        % child_accession
                    )
                    fetch_bioproject_children(
                        child_accession,
                        projects=projects,
                        children=children,
                        new_projects=new_projects,
                    )
    return children, new_projects


def fetch_accession(bioproject):
    """Fetch a GCA accession for a bioproject."""
    LOGGER.info("Fetching GCA accession for bioproject %s" % bioproject)
    url = "%s/search?result=assembly&query=study_accession%%3D%%22%s%%22&fields=accession%%2Cversion&format=tsv" % (
        ENA_API,
        bioproject,
    )
    result = tofetch.fetch_url(url)
    accession = None
    if result and result is not None:
        line = result.split("\n")[1]
        if line and "\t" in line:
            accession, version = line.split("\t")
            accession += ".%s" % version
    return accession


def fetch_bioproject_accessions(bioproject, *, projects=None):
    """Save assembly fasta file to local disk."""
    if projects is None:
        projects = []
    LOGGER.info("Fetching accessions under bioproject %s" % bioproject)
    accessions = []
    children, new_projects = fetch_bioproject_children(bioproject, projects=projects)
    project_list = set()
    for child_bioproject in children:
        sleep(0.5)
        accession = fetch_accession(child_bioproject)
        if accession is not None:
            accessions.append(accession)
            # project_list.add(child_bioproject)
            project_list.add(new_projects[child_bioproject])
    return accessions, project_list


if __name__ == "__main__":
    opts = docopt(__doc__)
    bioproject = opts["<BIOPROJECT>"]
    outdir = opts["--out"]
    os.makedirs(outdir, exist_ok=True)
    projects_file = "%s/bioprojects.processed" % outdir
    targets_file = "%s/accessions.todo" % outdir
    active_file = "%s/accessions.active" % outdir
    processed_file = "%s/accessions.processed" % outdir
    projects = tofile.read_file(projects_file)

    accessions, new_projects = fetch_bioproject_accessions(
        bioproject, projects=projects
    )
    with open(targets_file, "a+") as tfh:
        tfh.writelines([accession + "\n" for accession in accessions])
    with open(projects_file, "a+") as pfh:
        pfh.writelines([bioproject + "\n" for bioproject in new_projects])

    # TODO: introduce page numbering for when list gets long



#!/usr/bin/env python3

"""
BlobToolKit Pipeline.

Usage: blobtoolkit-pipeline [<command>] [<args>...] [-h|--help] [-v|--version]

Commands:
    data                      Fetch pipeline data
    run                       Run BlobToolKit Pipeline
    add-summary-to-metadata   Pipeline helper script
    chunk-fasta               Pipeline helper script
    count-busco-genes         Pipeline helper script
    extract-busco-genes       Pipeline helper script
    generate-config           Pipeline helper script
    generate-static-images    Pipeline helper script
    transfer-completed        Pipeline helper script
    unchunk-blast             Pipeline helper script
    window-stats              Pipeline helper script
    -h, --help      Show this
    -v, --version   Show version number
See 'blobtoolkit-pipeline <command> --help' for more information on a specific command.

"""

import sys

from docopt import DocoptExit
from docopt import docopt
from lib.version import __version__
from pkg_resources import working_set


def cli(rename=None):
    """Entry point."""
    # if sys.argv[1].match("-pipeline"):
    command = sys.argv[1]
    if len(sys.argv) > 2 and not sys.argv[2].startswith("-"):
        command = sys.argv[2]
    docs = __doc__
    if rename is not None:
        docs = docs.replace("blobtoolkit-pipeline", rename)
    try:
        args = docopt(docs, help=False, version=__version__)
    except DocoptExit:
        args = {"<command>": command}
    if args["<command>"]:
        # load <command> from entry_points
        for entry_point in working_set.iter_entry_points("blobtoolkit_pipeline.subcmd"):
            if entry_point.name == args["<command>"]:
                subcommand = entry_point.load()
                sys.exit(subcommand(rename))
    print(docs)
    raise DocoptExit



#!/usr/bin/env python3

"""
Generate config files for BlobToolKit pipeline.

Usage:
  blobtoolkit-pipeline generate-config <ACCESSION> [--coverage 30] [--download]
    [--out /path/to/output/directory] [--db /path/to/database/directory]
    [--db-suffix STRING] [--reads STRING...] [--read-runs INT] [--api-key STRING]
    [--platforms STRING] [--datasets] [--protocol STRING]
    [--download-client STRING] [--retry-times INT]

Options:
  --coverage=INT            Maximum coverage for read mapping [default: 30]
  --download                Flag to download remote files [default: False]
  --out PATH                Path to output directory [default: .]
  --db PATH                 Path to database directory [default: .]
  --db-suffix STRING        Database version suffix (e.g. 2021_06)
  --reads STRING            Read accession to include.
  --read-runs INT           Maximum number of read runs [default: 3]
  --api-key STRING          NCBI api key for use with edirect
  --platforms STRING        priority order for sequencing platforms
                            [default: PACBIO_SMRT,ILLUMINA_XTEN,ILLUMINA,OXFORD_NANOPORE,OTHER]
  --protocol STRING         Download files using ftp(s) or http(s) [default: ftp]
  --download-client STRING  Client to use for downloads (aria2c|curl|datasets) [default: curl]
  --retry-times INT         Number of times to retry a failed download [default: 0]
"""

# pyright: reportMissingModuleSource=false

import gzip
import os
import re
import sys
from glob import glob
from operator import itemgetter
from pathlib import Path
from shutil import rmtree
from shutil import which
from subprocess import PIPE
from subprocess import Popen
from subprocess import run
from zipfile import ZipFile

import requests
import ujson
import yaml
from defusedxml import ElementTree as ET
from docopt import DocoptExit
from docopt import docopt
from tolkein import tofetch
from tolkein import tofile
from tolkein import tolog

LOGGER = tolog.logger(__name__)

GOAT_API = "https://goat.genomehubs.org/api/v2"
ENA_API = "https://www.ebi.ac.uk/ena/browser/api"
GCA_NAME = re.compile(r"GCA_.+")
BUSCO_URL = "https://busco-data.ezlab.org/v5/data/lineages"


def find_busco_lineages(ancestors):
    """Work out which BUSCO sets to run for a given taxid."""
    LOGGER.info("Identifying relevant BUSCO lineages")
    BUSCO_SETS = {
        "422676": "aconoidasida",
        "7898": "actinopterygii",
        "5338": "agaricales",
        "155619": "agaricomycetes",
        "33630": "alveolata",
        "5794": "apicomplexa",
        "6854": "arachnida",
        "6656": "arthropoda",
        "4890": "ascomycota",
        "8782": "aves",
        "5204": "basidiomycota",
        "68889": "boletales",
        "3699": "brassicales",
        "134362": "capnodiales",
        "33554": "carnivora",
        "91561": "cetartiodactyla",
        "34395": "chaetothyriales",
        "3041": "chlorophyta",
        "5796": "coccidia",
        "28738": "cyprinodontiformes",
        "7147": "diptera",
        "147541": "dothideomycetes",
        "3193": "embryophyta",
        "33392": "endopterygota",
        "314146": "euarchontoglires",
        "33682": "euglenozoa",
        "2759": "eukaryota",
        "5042": "eurotiales",
        "147545": "eurotiomycetes",
        "9347": "eutheria",
        "72025": "fabales",
        "4751": "fungi",
        "314147": "glires",
        "1028384": "glomerellales",
        "5178": "helotiales",
        "7524": "hemiptera",
        "7399": "hymenoptera",
        "5125": "hypocreales",
        "50557": "insecta",
        "314145": "laurasiatheria",
        "147548": "leotiomycetes",
        "7088": "lepidoptera",
        "4447": "liliopsida",
        "40674": "mammalia",
        "33208": "metazoa",
        "6029": "microsporidia",
        "6447": "mollusca",
        "4827": "mucorales",
        "1913637": "mucoromycota",
        "6231": "nematoda",
        "33183": "onygenales",
        "9126": "passeriformes",
        "5820": "plasmodium",
        "92860": "pleosporales",
        "38820": "poales",
        "5303": "polyporales",
        "9443": "primates",
        "4891": "saccharomycetes",
        "8457": "sauropsida",
        "4069": "solanales",
        "147550": "sordariomycetes",
        "33634": "stramenopiles",
        "32523": "tetrapoda",
        "155616": "tremellomycetes",
        "7742": "vertebrata",
        "33090": "viridiplantae",
        "71240": "eudicots",
    }
    lineages = [
        f'{BUSCO_SETS[obj["taxon_id"]]}_odb10'
        for obj in ancestors
        if obj["taxon_id"] in BUSCO_SETS
    ]
    lineages.extend(("bacteria_odb10", "archaea_odb10"))
    return lineages


def fetch_file_curl(url, filename, protocol="ftp", retry_times=0):
    """fetch a remote file using curl."""
    filepath = Path(filename)
    if filepath.is_file():
        LOGGER.info("File exists, not overwriting")
        return
    if protocol:
        url = url.replace("ftp:", f"{protocol}:").replace("ftps:", f"{protocol}:")
    iteration = 0
    complete = False
    while iteration <= retry_times:
        cmd = ["curl", "-L", "-s", "--connect-timeout", "30"]
        if iteration:
            LOGGER.info("Connection interrupted, resuming download")
            cmd.extend(["-C", "-"])
        cmd.extend(
            [
                "-o",
                f"{filepath.parent}/{filepath.name}",
                url,
            ]
        )
        proc = run(cmd)
        if proc.returncode == 0:
            complete = True
            break
        iteration += 1
    if not complete:
        LOGGER.error(f"Unable to fetch {filename}")
        sys.exit(1)


def fetch_file_aria(url, filename, protocol="ftp", retry_times=0, binary="aria2c"):
    """fetch a remote file using aria2."""
    filepath = Path(filename)
    if filepath.is_file():
        LOGGER.info("File exists, not overwriting")
        return
    if protocol:
        url = url.replace("ftp:", f"{protocol}:").replace("ftps:", f"{protocol}:")
    iteration = 0
    complete = False
    while iteration <= retry_times:
        cmd = [binary, "-q", "-d", filepath.parent]
        if iteration:
            LOGGER.info("Connection interrupted, resuming download")
            cmd.append("-c")
        cmd.extend(["-o", filepath.name, url])
        proc = run(cmd)
        if proc.returncode == 0:
            complete = True
            break
        iteration += 1
    if not complete:
        LOGGER.error(f"Unable to fetch {filename}")
        sys.exit(1)


def fetch_ncbi_datasets_assembly(filename, accession, retry_times):
    """Fetch a remote assembly using NCBI datasets."""
    filepath = Path(filename)
    if filepath.is_file():
        LOGGER.info("File exists, not overwriting")
        return
    dataset_zip = f"{filepath.parents[0]}/{accession}_dataset.zip"
    download_dehydrated = [
        "datasets",
        "download",
        "genome",
        "accession",
        accession,
        "--dehydrated",
        "--filename",
        dataset_zip,
    ]
    proc = run(download_dehydrated)
    if proc.returncode == 0:
        LOGGER.error(f"Unable to fetch {accession} with NCBI datasets")
        sys.exit(1)
    with ZipFile(dataset_zip, "r") as zh:
        zh.extractall(filepath.parents[0])
    rehydrate = ["datasets", "rehydrate", "--directory", filepath.parents[0]]
    iteration = 0
    complete = False
    while iteration <= retry_times:
        proc = run(rehydrate)
        if proc.returncode == 0:
            complete = True
            break
        iteration += 1
    if not complete:
        LOGGER.error(f"Unable to rehydrate {accession} with NCBI datasets")
        sys.exit(1)
    assembly_dir = f"{filepath.parents[0]}/ncbi_dataset/data/{accession}/"
    LOGGER.info(f"Writing compressed file to {filename}")
    with open(glob(f"{assembly_dir}/*_genomic.fna")[0], "rb") as ifh:
        with gzip.open(filename, "wb") as ofh:
            ofh.writelines(ifh)
    os.remove(dataset_zip)
    os.remove(f"{filepath.parents[0]}/README.md")
    rmtree(f"{filepath.parents[0]}/ncbi_dataset")


def fetch_assembly_url(accession, api_key=None):
    """
    Fetch an assembly url using edirect.
    """
    eutils_env = os.environ.copy()
    if api_key and api_key is not None:
        eutils_env["NCBI_API_KEY"] = api_key
    esearch = Popen(
        f"esearch -db assembly -query {accession}",
        stdout=PIPE,
        shell=True,
        env=eutils_env,
    )

    esummary = Popen(
        "esummary", stdin=esearch.stdout, stdout=PIPE, shell=True, env=eutils_env
    )
    xtract = Popen(
        "xtract -pattern DocumentSummary -element FtpPath_GenBank",
        stdin=esummary.stdout,
        stdout=PIPE,
        shell=True,
    )
    xtract_stdout = xtract.communicate()[0].decode("utf-8").strip()
    for url in xtract_stdout.split("\n"):
        if basenames := re.findall(GCA_NAME, url):
            for basename in basenames:
                if accession in basename:
                    return f"{url}/{basename}_genomic.fna.gz"
    return None


def fetch_assembly_fasta(
    url, filename, accession, protocol, retry_times, download_client
):
    """Save assembly fasta file to local disk."""
    if download_client == "datasets" and which(download_client) is not None:
        LOGGER.info(f"Fetching assembly FASTA to {filename} using NCBI datasets")
        fetch_ncbi_datasets_assembly(filename, accession, retry_times)

    elif download_client.startswith("aria") and which(download_client) is not None:
        LOGGER.info(f"Fetching assembly FASTA to {filename} using {download_client}")
        fetch_file_aria(url, filename, protocol, retry_times, download_client)
    else:
        LOGGER.info(f"Fetching assembly FASTA to {filename} using curl")
        fetch_file_curl(url, filename, protocol, retry_times)


def write_cat_file(data, filename):
    result = [key for key, obj in data.items()]
    arr = [result]
    for idx, value in enumerate(data[result[0]]["list"]):
        row = [value]
        row.extend(data[key]["list"][idx] for key in result[1:])
        arr.append(row)
    tofile.write_file(filename, arr)
    return result


def parse_assembly_report(filename, cat_filename, syn_filename):
    """Parse synonyms and assembly level into tsv files."""
    cats = {
        "identifier": {"index": 4, "list": []},
        "assembly_role": {"index": 1, "list": []},
        "assembly_level": {"index": 3, "list": []},
        "assembly_unit": {"index": 7, "list": []},
    }
    names = {
        "identifier": {"index": 4, "list": []},
        "name": {"index": 0, "list": []},
        "assigned_name": {"index": 2, "list": []},
        "refseq_accession": {"index": 6, "list": []},
    }
    with tofile.open_file_handle(filename) as fh:
        for line in fh:
            if line.startswith("#"):
                continue
            row = line.rstrip().split("\t")
            for group in (cats, names):
                for obj in group.values():
                    value = row[obj["index"]]
                    obj["list"].append(value)
    write_cat_file(cats, cat_filename)
    write_cat_file(names, syn_filename)


def fetch_assembly_report(
    url, filename, cat_filename, syn_filename, protocol, retry_times
):
    """Save assembly report file to local disk."""
    LOGGER.info(f"Fetching assembly report to {filename}")
    fetch_file_curl(url, filename, protocol, retry_times)
    parse_assembly_report(filename, cat_filename, syn_filename)


def fetch_assembly_meta_xml(accession):
    """
    Fetch assembly metadata xml from ENA.
    """
    return tofetch.fetch_url(f"{ENA_API}/xml/{accession}")


def deep_find_text(data, tags):
    """
    Find nested attributes in xml.

    Return attribute value.
    """
    for tag in tags:
        try:
            data = data.find(tag)
        except Exception:
            return ""
    return data.text


def parse_assembly_meta(accession):
    """Return dict of metadata values for an assembly."""
    LOGGER.info("Fetching assembly metadata")
    meta = {
        "assembly": {"accession": accession},
        "busco": {"lineages": []},
        "reads": {"paired": [], "single": []},
        "revision": 0,
        "settings": {
            "tmp": "/tmp",
            "blast_chunk": 100000,
            "blast_max_chunks": 10,
            "blast_overlap": 0,
            "blast_min_length": 1000,
            "stats_chunk": 1000,
            "stats_windows": [0.1, 0.01, 100000, 1000000],
        },
        "similarity": {
            "defaults": {
                "evalue": 1.0e-10,
                "import_evalue": 1.0e-25,
                "max_target_seqs": 10,
                "taxrule": "buscogenes",
            },
            "diamond_blastx": {"name": "reference_proteomes"},
            "diamond_blastp": {
                "name": "reference_proteomes",
                "import_max_target_seqs": 100000,
                "taxrule": "blastp=buscogenes",
            },
            "blastn": {"name": "nt"},
        },
        "taxon": {},
        "version": 1,
    }
    xml = fetch_assembly_meta_xml(accession)
    try:
        root = ET.fromstring(xml)
    except TypeError:
        LOGGER.error(f"Invalid accession: '{accession}'")
        sys.exit(1)
    asm = root.find("ASSEMBLY")
    meta["assembly"]["bioproject"] = deep_find_text(
        asm, ("STUDY_REF", "IDENTIFIERS", "PRIMARY_ID")
    )
    meta["assembly"]["biosample"] = deep_find_text(
        asm, ("SAMPLE_REF", "IDENTIFIERS", "PRIMARY_ID")
    )
    meta["taxon"]["taxid"] = deep_find_text(asm, ("TAXON", "TAXON_ID"))
    meta["taxon"]["name"] = deep_find_text(asm, ("TAXON", "SCIENTIFIC_NAME"))
    meta["assembly"]["level"] = asm.find("ASSEMBLY_LEVEL").text
    meta["assembly"]["alias"] = asm.attrib["alias"]
    wgs_prefix = deep_find_text(asm, ("WGS_SET", "PREFIX"))
    wgs_version = deep_find_text(asm, ("WGS_SET", "VERSION"))
    if wgs_prefix and wgs_version:
        meta["assembly"]["prefix"] = f"{wgs_prefix}{wgs_version.zfill(2)}"
    elif " " not in meta["assembly"]["alias"]:
        meta["assembly"]["prefix"] = meta["assembly"]["alias"].replace(".", "_")
    else:
        meta["assembly"]["prefix"] = meta["assembly"]["accession"].replace(".", "_")
    attributes = asm.find("ASSEMBLY_ATTRIBUTES")
    for attribute in attributes.findall("ASSEMBLY_ATTRIBUTE"):
        if attribute.find("TAG").text == "total-length":
            meta["assembly"]["span"] = int(attribute.find("VALUE").text)
        elif attribute.find("TAG").text == "scaffold-count":
            meta["assembly"]["scaffold-count"] = int(attribute.find("VALUE").text)
    return meta


def fetch_busco_lineages(busco_sets, buscodir):
    """Fetch busco lineages."""
    if not busco_sets:
        return
    lineages_to_fetch = []
    for lineage in busco_sets:
        busco_lineage = f"{buscodir}/lineages/{lineage}"
        if not os.path.isdir(busco_lineage):
            lineages_to_fetch.append(lineage)
    if not lineages_to_fetch:
        return
    lineage_urls = {}
    LOGGER.info("Fetching BUSCO lineage directory listing")
    listing = tofetch.fetch_url(f"{BUSCO_URL}/")
    for entry in listing.split("\n"):
        parts = re.split(r"[\"\s]+", entry)
        if len(parts) == 8:
            busco_set = re.sub(r"\..+$", "", parts[2])
            lineage_urls[busco_set] = f"{BUSCO_URL}/{parts[2]}"
    for lineage in lineages_to_fetch:
        LOGGER.info(f"Fetching BUSCO lineage {lineage}")
        tofetch.fetch_tar(lineage_urls[lineage], buscodir)


def fetch_goat_data(taxon_id):
    """Fetch taxon metadata from GoaT."""
    LOGGER.info("Fetching taxon metadata")
    url = f"{GOAT_API}/record?recordId=taxon_id-{taxon_id}&result=taxon"
    result = tofetch.fetch_url(url)
    if result is None:
        LOGGER.error("Unable to fetch taxon metadata for '%s' from GoaT", taxon_id)
        sys.exit(1)
    data = ujson.loads(result)
    return data["records"][0]["record"]


def fetch_read_info(accession, per_platform):
    """Fetch read info for an accession."""
    portal = "https://www.ebi.ac.uk/ena/portal/api"
    url = f"{portal}/filereport?accession={accession}&result=read_run&fields=run_accession,fastq_bytes,base_count,library_strategy,library_selection,library_layout,instrument_platform,experiment_title,fastq_ftp"

    si_suffix = {
        "k": 3,
        "M": 6,
        "G": 9,
        "T": 12,
        "P": 15,
        "E": 18,
    }

    data = tofetch.fetch_url(url)
    if data is None:
        return
    header = None
    for line in data.split("\n"):
        if not line or line == "":
            continue
        if header is None:
            header = line.split("\t")
            continue
        fields = line.split("\t")
        values = {}
        platform = "OTHER"
        for i in range(len(header)):
            value = fields[i]
            if header[i] == "instrument_platform" and platform != "ILLUMINA_XTEN":
                platform = fields[i]
            if (
                header[i] == "experiment_title"
                and value == "HiSeq X Ten paired end sequencing"
            ):
                platform = "ILLUMINA_XTEN"
            values[header[i]] = value
        try:
            values["base_count"] = int(values["base_count"])
        except ValueError:
            if "base_count" in values and values["base_count"].endswith("b"):
                values["base_count"] = values["base_count"].strip("b")
                suffix = values["base_count"][-1]
                try:
                    exponent = si_suffix[suffix]
                except KeyError:
                    exponent = 0
                values["base_count"] = int(values["base_count"].strip(suffix)) * pow(
                    10, exponent
                )
            else:
                values["base_count"] = 0
        except KeyError:
            values["base_count"] = 0
        try:
            per_platform[platform].append(values)
        except KeyError:
            per_platform["OTHER"].append(values)


def assembly_reads(accession, read_runs, platforms):
    """
    Query INSDC reads for an <accession> (or list of accessions).

    Return a dict of SRA accession, FASTQ ftp url, md5 and file size.
    """
    sra = []
    per_platform = {
        "PACBIO_SMRT": [],
        "ILLUMINA_XTEN": [],
        "ILLUMINA": [],
        "OXFORD_NANOPORE": [],
        "OTHER": [],
    }
    if not isinstance(accession, list):
        accession = [accession]
    for acc in accession:
        fetch_read_info(acc, per_platform)
    for key in platforms.split(","):
        runs = []
        if arr := per_platform[key]:
            runs.extend(iter(arr))
        runs = sorted(runs, key=itemgetter("base_count"), reverse=True)[
            : read_runs - len(sra)
        ]
        sra += runs
        if len(sra) >= read_runs:
            break
    return sra or None


def base_count(x):
    """Return number of bases or zero."""
    return int(x["base_count"][0] or 0) if isinstance(x["base_count"], list) else 0


def fetch_read_files(meta, protocol, retry_times):
    """Fetch sra reads."""
    files = meta["file"].split(";")
    for index, url in enumerate(meta["fastq_ftp"].split(";")):
        url = f"ftp://{url}"
        read_file = files[index]
        LOGGER.info("Fetching read file %s", read_file)
        fetch_file_curl(url, read_file, protocol, retry_times)


def add_taxon_to_meta(meta, taxon_meta):
    """Add taxon info to metadata."""
    LOGGER.info("Adding taxon metadata to assembly metadata")
    ranks = [
        "species",
        "genus",
        "family",
        "order",
        "class",
        "phylum",
        "kingdom",
        "superkingdom",
    ]
    for obj in taxon_meta["lineage"]:
        if obj["taxon_rank"] in ranks:
            meta["taxon"].update({obj["taxon_rank"]: obj["scientific_name"]})


def add_reads_to_meta(meta, sra, readdir):
    """Add read accessions to metadata."""
    LOGGER.info("Adding read accessions to assembly metadata")
    for library in sra:
        strategy = library["library_layout"].lower()
        fastq_ftp = library["fastq_ftp"].split(";")[-2:]
        info = {
            "prefix": library["run_accession"],
            "platform": library["instrument_platform"],
            "base_count": library["base_count"],
            "file": ";".join(
                [re.sub(r"^.+/", f"{readdir}/", url) for url in fastq_ftp]
            ),
            "url": ";".join(fastq_ftp),
        }

        library["file"] = info["file"]
        meta["reads"][strategy].append(info)


def set_btk_version(meta):
    """Get curent version of hosted datasets from BTK API."""
    LOGGER.info("Checking current version on BTK public viewer")
    string = meta["assembly"]["prefix"]
    btk = f"https://blobtoolkit.genomehubs.org/api/v1/search/{string}"
    response = requests.get(btk)
    current = 0
    if response.ok:
        data = yaml.full_load(response.text)
        for asm in data:
            current = asm["version"] if "version" in asm else 1
    meta["revision"] = current
    meta["version"] = current + 1


def process_reads(opts, outdir, protocol, retry_times, meta, sra):
    if opts["--coverage"]:
        meta["reads"].update({"coverage": {"max": int(opts["--coverage"])}})
    readdir = f"{outdir}/reads"
    add_reads_to_meta(meta, sra, readdir)
    if opts["--download"]:
        os.makedirs(readdir, exist_ok=True)
        for library in sra:
            fetch_read_files(library, protocol, retry_times)


def process_busco(opts, buscodir, meta, taxon_meta):
    busco_sets = find_busco_lineages(taxon_meta["lineage"])
    if busco_sets:
        meta["busco"].update(
            {
                "download_dir": buscodir,
                "lineages": busco_sets,
                "basal_lineages": [
                    "eukaryota_odb10",
                    "bacteria_odb10",
                    "archaea_odb10",
                ],
            }
        )
    if opts["--download"]:
        fetch_busco_lineages(busco_sets, buscodir)


def download_assembly(
    accession,
    outdir,
    protocol,
    download_client,
    retry_times,
    buscodir,
    assembly_url,
    assembly_file,
    assembly_report,
    syn_filename,
    cat_filename,
):
    os.makedirs(buscodir, exist_ok=True)
    os.makedirs(f"{outdir}/assembly", exist_ok=True)
    fetch_assembly_fasta(
        assembly_url,
        assembly_file,
        accession,
        protocol,
        retry_times,
        download_client,
    )
    report_url = assembly_url.replace("_genomic.fna.gz", "_assembly_report.txt")
    fetch_assembly_report(
        report_url,
        assembly_report,
        cat_filename,
        syn_filename,
        protocol,
        retry_times,
    )


def set_defaults(opts):
    accession = opts["<ACCESSION>"]
    outdir = opts["--out"]
    dbdir = opts["--db"]
    protocol = opts["--protocol"]
    download_client = opts["--download-client"]
    retry_times = int(opts["--retry-times"])
    buscodir = f"{dbdir}/busco"
    uniprotdir = f"{dbdir}/uniprot"
    ntdir = f"{dbdir}/nt"
    taxdumpdir = f"{dbdir}/taxdump"
    if opts["--db-suffix"]:
        buscodir += f'_{opts["--db-suffix"]}'
        ntdir += f'_{opts["--db-suffix"]}'
        uniprotdir += f'_{opts["--db-suffix"]}'
        taxdumpdir += f'_{opts["--db-suffix"]}'
    if not outdir.endswith(accession):
        outdir += f"/{accession}"
    return (
        accession,
        outdir,
        protocol,
        download_client,
        retry_times,
        buscodir,
        uniprotdir,
        ntdir,
        taxdumpdir,
    )


def main(rename=None):
    """Entry point."""
    docs = __doc__
    if rename is not None:
        docs = docs.replace("blobtoolkit-pipeline", rename)
    try:
        opts = docopt(docs)
    except DocoptExit as e:
        raise DocoptExit from e
    (
        accession,
        outdir,
        protocol,
        download_client,
        retry_times,
        buscodir,
        uniprotdir,
        ntdir,
        taxdumpdir,
    ) = set_defaults(opts)
    os.makedirs(outdir, exist_ok=True)
    meta = parse_assembly_meta(accession)
    assembly_url = fetch_assembly_url(accession, opts["--api-key"])
    if assembly_url is None:
        LOGGER.error("Unable to find assembly URL")
        sys.exit(1)
    assembly_file = f"{outdir}/assembly/{accession}.fasta.gz"
    meta["assembly"].update({"file": assembly_file, "url": assembly_url})
    assembly_report = f"{outdir}/assembly/{accession}.report.txt"
    syn_filename = f"{outdir}/assembly/{accession}.synonyms.tsv"
    cat_filename = f"{outdir}/assembly/{accession}.categories.tsv"
    meta["fields"] = {
        "synonyms": {"file": syn_filename, "prefix": "insdc"},
        "categories": {"file": cat_filename},
    }
    if opts["--download"]:
        download_assembly(
            accession,
            outdir,
            protocol,
            download_client,
            retry_times,
            buscodir,
            assembly_url,
            assembly_file,
            assembly_report,
            syn_filename,
            cat_filename,
        )
    taxon_meta = fetch_goat_data(meta["taxon"]["taxid"])
    add_taxon_to_meta(meta, taxon_meta)
    set_btk_version(meta)
    process_busco(opts, buscodir, meta, taxon_meta)
    read_accessions = []
    if meta["assembly"]["biosample"]:
        read_accessions = [meta["assembly"]["biosample"]]
    if opts["--reads"]:
        read_accessions += opts["--reads"]
    if sra := assembly_reads(
        read_accessions, int(opts["--read-runs"]), opts["--platforms"]
    ):
        process_reads(opts, outdir, protocol, retry_times, meta, sra)
    meta["similarity"]["blastn"]["path"] = ntdir
    meta["similarity"]["diamond_blastx"]["path"] = uniprotdir
    meta["similarity"]["diamond_blastp"]["path"] = uniprotdir
    meta["settings"]["taxdump"] = taxdumpdir
    tofile.write_file(f"{outdir}/config.yaml", meta)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
Fetch BlobToolKit Pipeline data.

NOT YET IMPLEMENTED

Usage: blobtoolkit-pipeline data --config YAML

Options:
    --config YAML  YAML format configuration filename.
"""

import logging
import re
import sys
from collections import defaultdict

from docopt import DocoptExit
from docopt import docopt

logger_config = {
    "level": logging.INFO,
    "format": "%(asctime)s [%(levelname)s] line %(lineno)d %(message)s",
    "filemode": "w",
}
logging.basicConfig(**logger_config)
logger = logging.getLogger()


def main(rename=None):
    """Entry point."""
    docs = __doc__
    if rename is not None:
        docs = docs.replace("blobtoolkit-pipeline", rename)
    try:
        args = docopt(docs)
    except DocoptExit as e:
        raise DocoptExit from e



#!/usr/bin/env python3

"""
Add summary data to metadata.

Usage: blobtoolkit-pipeline add-summary-to-metadata --config FILE --out FILE

Options:
    --config FILE   YAML format config file
    --out FILE      Output file
"""


import logging
import re
import subprocess
import traceback
from collections import OrderedDict

import yaml
from docopt import DocoptExit
from docopt import docopt
from tolkein import tofile

from .functions import read_similarity_settings
from .functions import reads_by_prefix
from .version import __version__

logger_config = {
    "level": logging.INFO,
    "format": "%(asctime)s [%(levelname)s] line %(lineno)d %(message)s",
    "filemode": "w",
}
logging.basicConfig(**logger_config)
logger = logging.getLogger()

logger.info(f"Starting script: {__file__}")


def add_pipeline_version(meta):
    """Add pipeline version info to metadata."""
    meta["settings"]["pipeline"] = "https://github.com/blobtoolkit/blobtoolkit"
    meta["settings"]["release"] = __version__


def add_software_versions(meta):
    """Add software versions to meta."""
    programs = {
        "blastn": {"flag": "-version", "regex": r":\s*(\S+)$"},
        "blobtools": {"regex": r"\s*v(\S+)$"},
        "busco": {"regex": r"\s+(\S+)$"},
        "diamond": {"regex": r"version\s*(\S+)$"},
        "minimap2": {"regex": r"(\S+)"},
        "blobtk": {"regex": r"\s+(\S+)$"},
        "python": {"regex": r"\s+(\S+)$"},
        "samtools": {"regex": r"\s+(\S+)$"},
        "seqtk": {"flag": "", "line": 2, "status": 1, "regex": r":\s*(\S+)$"},
        "snakemake": {"regex": r"(\S+)"},
    }
    versions = {}
    logger.info("Adding software version details")
    for key, opts in programs.items():
        cmd = [key]
        if flag := opts.get("flag", "--version"):
            cmd.append(flag)
        try:
            p = subprocess.Popen(
                cmd,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                encoding="utf-8",
            )
            (output, err) = p.communicate()
            status = p.wait()
            expected_status = opts.get("status", 0)
            if expected_status == 1:
                output = err
                status = 0
            if status == 0:
                for idx, line in enumerate(output.split("\n")):
                    if idx == opts.get("line", 0):
                        match = re.search(opts["regex"], line)
                        version = match.groups()[0]
            else:
                version = "unknown"
        except FileNotFoundError:
            version = "not found"
        logger.info("%s: %s", key, version)
        versions[key] = version
    meta["settings"]["software_versions"] = versions


def main(rename=None):
    """Entry point."""
    docs = __doc__
    if rename is not None:
        docs = docs.replace("blobtoolkit-pipeline", rename)
    try:
        args = docopt(docs)
    except DocoptExit as e:
        raise DocoptExit from e
    try:
        meta = {}
        config = args["--config"]
        if not isinstance(config, dict):
            config = tofile.load_yaml(config)
        meta["assembly"] = config["assembly"]
        meta["taxon"] = config["taxon"]
        meta["settings"] = {
            "blast_chunk": 100000,
            "blast_max_chunks": 10,
            **config["settings"],
        }
        meta["similarity"] = {
            "diamond_blastx": read_similarity_settings(config, "diamond_blastx"),
            "diamond_blastp": read_similarity_settings(config, "diamond_blastp"),
        }
        meta["reads"] = reads_by_prefix(config)
        for key, value in meta["reads"].items():
            if "url" in value and not isinstance(value["url"], list):
                value["url"] = value["url"].split(";")
            elif "url" not in value:
                value["url"] = []
        add_pipeline_version(meta)
        add_software_versions(meta)
        meta["version"] = config.get("version", 1)
        meta["revision"] = config.get("revision", 0)

        yaml.add_representer(
            OrderedDict,
            lambda dumper, data: dumper.represent_mapping(
                "tag:yaml.org,2002:map", data.items()
            ),
        )
        yaml.add_representer(
            tuple,
            lambda dumper, data: dumper.represent_sequence(
                "tag:yaml.org,2002:seq", data
            ),
        )
        yaml.Dumper.ignore_aliases = lambda *args: True
        with open(args["--out"], "w") as fh:
            fh.write(yaml.dump(meta, indent=1))

    except Exception as err:
        logger.error(traceback.format_exc())
        exit(13)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

"""
Restore intermediate files from partial analyses.

Usage:
  blobtoolkit-pipeline resume-pipeline --in PATH [--update-cov] --out PATH

Options:
  --in PATH              Path to input directory.
  --update-cov           Flag to discard existing coverage data.
  --out PATH             Path to output directory.
"""

import glob
import gzip
import re
import shutil
import tarfile
import zlib
from pathlib import Path
from time import sleep

import yaml
from docopt import DocoptExit
from docopt import docopt
from tolkein import tofile
from tolkein import tolog

LOGGER = tolog.logger(__name__)


def stream_gzip_decompress(stream):
    dec = zlib.decompressobj(32 + zlib.MAX_WBITS)  # offset 32 to skip the header
    for chunk in stream:
        rv = dec.decompress(chunk)
        if rv:
            yield rv


def untar_directory(archive, destdir, *, update_cov=False, compress=False):
    LOGGER.info("Extracting files from %s to %s", archive, destdir)
    # make sure destdir exists
    Path(destdir).mkdir(parents=True, exist_ok=True)
    # extract files
    with tarfile.open(archive) as tfh:
        for member in tfh.getmembers():
            if ".stats." in member.name:
                f = tfh.extractfile(member)
                name = re.sub(r"^.+\/", "", member.name)
                name = name.replace(".gz", "")
                with open("%s/%s" % (destdir, name), "wb") as ofh:
                    for chunk in stream_gzip_decompress(f):
                        ofh.write(chunk)
    tools = [
        "windowmasker",
        "chunk_stats",
        "busco",
        "minimap",
        "cov_stats",
        "diamond",
        "diamond_blastp",
        "blastn",
        "window_stats",
        "blobtools",
        "view",
    ]
    if update_cov:
        tools = tools[0:3] + tools[5:8]
    for tool in tools:
        stats_file = Path("%s/%s.stats" % (destdir, tool))
        if stats_file.is_file():
            stats_file.touch()
            sleep(1)


def list_accessions(config):
    """List assembly and read accessions."""
    try:
        gca = config["assembly"]["accession"]
    except KeyError:
        return None
    try:
        paired = [obj["prefix"] for obj in config["reads"]["paired"]]
    except KeyError:
        paired = []
    try:
        single = [obj["prefix"] for obj in config["reads"]["single"]]
    except KeyError:
        single = []
    return f"{gca} {' '.join(paired+single)}"


def main(rename=None):
    """Entry point."""
    docs = __doc__
    if rename is not None:
        docs = docs.replace("blobtoolkit-pipeline", rename)
    try:
        opts = docopt(docs)
    except DocoptExit as e:
        raise DocoptExit from e
    indir = opts["--in"]
    config_file = "%s/config.yaml" % indir
    data = tofile.read_file(config_file)
    config = yaml.load(data, Loader=yaml.BaseLoader)
    destdir = opts["--out"]
    accession = config["assembly"]["accession"]
    Path("%s/%s" % (destdir, accession)).mkdir(parents=True, exist_ok=True)
    shutil.copy2(config_file, "%s/%s/config.yaml" % (destdir, accession))
    tools = [
        "busco",
        "diamond",
        "diamond_blastp",
        "blastn",
        "chunk_stats",
        "window_stats",
        "windowmasker",
    ]
    for tool in tools:
        Path("%s/%s/%s" % (destdir, accession, tool)).mkdir(parents=True, exist_ok=True)
        for gzfile in glob.glob("%s/*%s*.gz" % (indir, tool)):
            with gzip.open(gzfile, "rb") as fh:
                file = gzfile.replace(".gz", "")
                file = file.replace(
                    "diamond_blastp.busco_genes.out", "diamond.busco_genes.out"
                )
                file = file.replace(
                    "diamond_blastx.reference_proteomes.out",
                    "diamond.reference_proteomes.out",
                )
                file = file.replace(indir, "%s/%s/%s" % (destdir, accession, tool))
                if tool == "chunk_stats":
                    file = file.replace(".tsv", ".mask.bed")
                    with open(file, "w") as ofh:
                        next(fh)
                        for line in fh:
                            ofh.write(
                                "\t".join(line.decode("utf-8").split("\t")[:3]) + "\n"
                            )
                else:
                    with open(file, "wb") as ofh:
                        shutil.copyfileobj(fh, ofh)

    for tf in glob.glob("%s/*.pipeline.tar" % indir):
        untar_directory(
            tf, "%s/%s" % (destdir, accession), update_cov=opts["--update-cov"]
        )
    for tf in glob.glob("%s/*.busco*.tar" % indir):
        with tarfile.open(tf) as tfh:
            tfh.extractall(path="%s/%s/busco" % (destdir, accession))
    print(list_accessions(config))


if __name__ == "__main__":
    main()



import pathlib


def reads_by_prefix(config):
    """Return read meta by prefix"""
    reads = {}
    if "reads" not in config:
        return {}

    for strategy in ("paired", "single"):
        if not strategy in config["reads"] or not config["reads"][strategy]:
            continue
        for entry in config["reads"][strategy]:
            if isinstance(entry, list):
                meta = {
                    "prefix": entry[0],
                    "platform": entry[1],
                    "strategy": strategy,
                    "base_count": entry[2],
                    "file": entry[3],
                }
                if len(entry) == 5:
                    meta["url"] = entry[4].split(";")
            else:
                meta = entry
            reads.update({meta["prefix"]: meta})
    return reads


def minimap_tuning(config, prefix):
    """Set minimap2 mapping parameter."""
    reads = reads_by_prefix(config)
    tunings = {
        "ILLUMINA": "sr",
        "OXFORD_NANOPORE": "map-ont",
        "PACBIO_SMRT": "map-pb",
        "OTHER": "sr",
    }
    return tunings.get(reads[prefix]["platform"], "sr")


def read_files(config, prefix):
    """Get read filenames."""
    reads = reads_by_prefix(config)
    return reads[prefix]["file"].split(";")


def seqtk_sample_input(config, prefix):
    """Generate seqtk command to subsamplereads if required."""
    meta = reads_by_prefix(config)[prefix]
    filenames = meta["file"].split(";")
    ratio = 1
    if "coverage" in config["reads"] and "max" in config["reads"]["coverage"]:
        base_count = meta.get("base_count", None)
        assembly_span = config["assembly"].get("span", None)
        if isinstance(base_count, int) and isinstance(assembly_span, int):
            try:
                ratio = assembly_span * config["reads"]["coverage"]["max"] / base_count
            except ZeroDivisionError:
                pass
    if ratio <= 0.95:
        command = " ".join(
            [
                "<(seqtk sample -s 100 %s %.2f)" % (filename, ratio)
                for filename in filenames
            ]
        )
    else:
        command = " ".join(filenames)
    return command


def diamond_db_name(config):
    """Generate filtered diamond database name."""
    name = "reference_proteomes"
    parts = ["diamond", name]
    return ".".join(parts)


def blobdir_name(config):
    """Generate blobdir name."""
    name = config["assembly"]["prefix"]
    if "revision" in config and config["revision"] > 0:
        name = "%s.%d" % (name, config["revision"])
    return name


def blobtools_cov_flag(config):
    """Generate --cov flag for blobtools add."""
    keys = reads_by_prefix(config).keys()
    if keys:
        return "--cov " + " --cov ".join(
            [
                "%s/%s.%s.bam=%s"
                % (minimap_path, config["assembly"]["prefix"], key, key)
                for key in keys
            ]
        )
    return ""


def gzipped_bed_cols(config):
    """Generate cut commands for gzipped bed files."""
    keys = reads_by_prefix(config).keys()
    if keys:
        return " ".join(
            [
                "<(echo %s_cov && gunzip -c %s.%s.regions.bed.gz | cut -f 4)"
                % (key, config["assembly"]["prefix"], key)
                for key in keys
            ]
        )
    return ""


def set_stats_chunk(config):
    """Set chunk size for calculating stats."""
    return config["settings"].get("stats_chunk", 1000)


def skip_windowmasker(config):
    """Set flag to skip windowmasker step."""
    return config["settings"].get("skip_windowmasker", False)


def set_stats_windows(config):
    """Set window sizes for averaging stats."""
    windows = config["settings"].get("stats_windows", 0.1)
    if not isinstance(windows, list):
        windows = [windows]
    if 1 not in windows:
        windows.append(1)
    return windows


def set_blast_chunk(config):
    """Set minimum chunk size for splitting long sequences."""
    return config["settings"].get("blast_chunk", 100000)


def set_blast_chunk_overlap(config):
    """Set overlap length for splitting long sequences."""
    return config["settings"].get("blast_overlap", 0)


def set_blast_max_chunks(config):
    """Set minimum chunk size for splitting long sequences."""
    return config["settings"].get("blast_max_chunks", 10)


def set_blast_min_length(config):
    """Set minimum sequence length for running blast searches."""
    return config["settings"].get("blast_min_length", 1000)


def taxid_flag(config, tool):
    """Set taxid flag for running blast searches."""
    taxid = config["taxon"].get("taxid", None)
    if taxid is None:
        return ""
    flags = {
        "blastn": "-negative_taxids",
        "diamond_blastp": "--taxon-exclude",
        "diamond_blastx": "--taxon-exclude",
    }
    return "%s %s" % (flags[tool], taxid)


def read_similarity_settings(config, group):
    """Read similarity settings for blast rules and outputs."""
    settings = {
        "evalue": 1.0e-10,
        "import_evalue": 1.0e-25,
        "max_target_seqs": 10,
        "name": "reference_proteomes",
    }
    if "defaults" in config["similarity"]:
        settings.update({**config["similarity"]["defaults"]})
    if group in config["similarity"]:
        settings.update({**config["similarity"][group]})
    return settings


def similarity_setting(config, group, value):
    """Get a single similarity setting value."""
    settings = read_similarity_settings(config, group)
    setting = settings.get(value, None)
    if setting is not None:
        return setting
    settings = {
        "evalue": 1.0e-25,
        "max_target_seqs": 10,
        "taxrule": "buscogenes",
        **settings,
    }
    if value.startswith("import_"):
        value = value.replace("import_", "")
    return settings[value]


def set_update_taxrule(config):
    taxrule = similarity_setting(config, "diamond_blastx", "taxrule")
    if taxrule == "buscoregions":
        return "--update-plot"
    return ""


def set_fields(config):
    params = ""
    if "fields" in config and isinstance(config["fields"], dict):
        for key, obj in config["fields"].items():
            if "file" in obj:
                if key == "synonyms":
                    prefix = obj.get("prefix", "")
                    if prefix:
                        prefix = "=%s" % prefix
                    params += " --synonyms %s%s" % (obj["file"], prefix)
                else:
                    params += " --text %s" % (obj["file"])
    if params:
        params += " --text-header"
    return params


def get_basal_lineages(config):
    """Get basal BUSCO lineages from config."""
    if "basal_lineages" in config["busco"]:
        return config["busco"]["basal_lineages"]
    basal = {"archaea_odb10", "bacteria_odb10", "eukaryota_odb10"}
    lineages = []
    if "lineages" in config["busco"]:
        for lineage in config["busco"]["lineages"]:
            if lineage in basal:
                lineages.push(lineage)
    return lineages


def set_view_timeout(config):
    """Set timeout duration for generating static images."""
    return config["settings"].get("view_timeout", 60)



#!/usr/bin/env python3
"""
Count BUSCO genes.

Usage: blobtoolkit-pipeline count-busco-genes --in TSV... --mask TSV --out TSV

Options:
    --in TSV      BUSCO full table tsv file.
    --mask TSV    BED or BED-like TSV format mask file to specify sequence chunks.
    --out TSV     output TSV filename or suffix.
"""

import logging
from collections import defaultdict

from docopt import DocoptExit
from docopt import docopt
from tolkein import tofile

logger_config = {
    "level": logging.INFO,
    "format": "%(asctime)s [%(levelname)s] line %(lineno)d %(message)s",
    "filemode": "w",
}
logging.basicConfig(**logger_config)
logger = logging.getLogger()


def load_mask(filename):
    """Load bed file as mask."""
    mask = defaultdict(dict)
    header = []
    with tofile.open_file_handle(filename) as fh:
        for line in fh.readlines():
            seqid, start, end, *cols = line.rstrip().split("\t")
            if cols is None:
                cols = []
            if seqid == "sequence" and start == "start":
                header = cols
                continue
            mask[seqid].update({int(start): {"end": int(end), "cols": cols}})
    return mask, header


def parse_busco_summary(filename, mask, header):
    """Parse chunked values into dict."""
    lineage = None
    with tofile.open_file_handle(filename) as fh:
        buscos = defaultdict(list)
        for line in fh.readlines():
            if line.startswith("#"):
                if line.startswith("# The lineage dataset is:"):
                    meta = line.split()
                    lineage = meta[5]
                    header.append(f"{lineage}_count")
                continue
            _, status, *rest = line.rstrip().split("\t")
            if status in {"Fragmented", "Missing"}:
                continue
            seqid, start, end, strand, *rest = rest
            if ":" in seqid and strand == "-":
                start = end
            buscos[seqid.split(":")[0]].append(int(start))
    if lineage is not None:
        for seqid in mask:
            starts = sorted(buscos[seqid])
            i = 0
            for start, obj in mask[seqid].items():
                ctr = 0
                while i < len(starts):
                    if starts[i] >= start:
                        if starts[i] > obj["end"]:
                            break
                        ctr += 1
                    i += 1
                obj["cols"].append(ctr)
    return mask, header


def main(rename=None):
    """Entry point."""
    docs = __doc__
    if rename is not None:
        docs = docs.replace("blobtoolkit-pipeline", rename)
    try:
        args = docopt(docs)
    except DocoptExit as e:
        raise DocoptExit from e
    try:
        mask, header = load_mask(args["--mask"])
        for buscofile in args["--in"]:
            mask, header = parse_busco_summary(buscofile, mask, header)
        outfile = args["--out"]
        if outfile.startswith("."):
            outfile = "%s%s" % (args["--in"], args["--out"])
        header = ["sequence", "start", "end"] + header
        rows = ["%s\n" % "\t".join(header)]
        for seqid in mask:
            for start, obj in mask[seqid].items():
                row = [seqid, start, obj["end"]] + obj["cols"]
                rows.append("%s\n" % "\t".join([str(v) for v in row]))
        with open(outfile, "w") as ofh:
            ofh.writelines(rows)
    except Exception as err:
        raise err


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

"""
Generate static image files.

Usage:
  blobtoolkit-pipeline generate-static-images --blobdir STRING [--coverage] [--host STRING]
                         [--ports RANGE] [--timeout INT] [--path PATH]

Options:
  --blobdir STRING  BlobDir to generate static views for.
  --coverage        Flag to indicate if coverage data are available and blob views should be generated. [Default False].
  --host STRING     Host name. [Default: http://localhost]
  --path STRING     Path to directory containing BlobDir. [Default: .]
  --ports RANGE     Range of ports available for hosting BlobTools Viewer/API. [Default: 8000-8099] = str(snakemake.params.ports)
  --timeout INT     Maximum time (in seconds) to wait for images. [Default: 30]
"""

import glob
import logging
import os
import shlex
import signal
import subprocess
import sys
from pathlib import Path

from docopt import DocoptExit
from docopt import docopt

logger_config = {
    "level": logging.INFO,
    "format": "%(asctime)s [%(levelname)s] line %(lineno)d %(message)s",
    "filemode": "w",
}
logging.basicConfig(**logger_config)
logger = logging.getLogger()


# def parse_args():
#     """Parse snakemake args if available."""
#     args = {}
#     try:
#         args["--blobdir"] = snakemake.wildcards.blobdir
#         args["--host"] = str(snakemake.params.host)
#         args["--ports"] = str(snakemake.params.ports)
#         args["--timeout"] = str(snakemake.params.timeout)
#         for key, value in args.items:
#             sys.argv.append(key)
#             sys.argv.append(value)
#         if snakemake.input.cov:
#             args["--coverage"] = True
#     except NameError as err:
#         pass


def main(rename=None):
    """Entry point."""
    docs = __doc__
    if rename is not None:
        docs = docs.replace("blobtoolkit-pipeline", rename)
    try:
        args = docopt(docs)
    except DocoptExit as e:
        raise DocoptExit from e

    blob_path = "%s/%s" % (args["--path"], args["--blobdir"])

    try:
        views = [
            "--view blob --param plotShape=circle --param largeFonts=true --format png",
            "--view blob --param plotShape=hex --param largeFonts=true --format png",
            "--view blob --param plotShape=square --param largeFonts=true --format png",
            "--view blob --param plotShape=kite --param largeFonts=true --format png",
            "--view cumulative --param largeFonts=true --format png",
            "--view snail --param largeFonts=true --format png",
        ]

        if not args["--coverage"]:
            views = views[4:]

        cmds = []

        for view in views:
            cmds.append(
                "blobtools view --host %s --timeout %s --ports %s %s --out %s/ %s"
                % (
                    args["--host"],
                    args["--timeout"],
                    args["--ports"],
                    view,
                    blob_path,
                    args["--blobdir"],
                )
            )

        cmds.append(
            "blobtools filter --summary %s/%s.summary.json %s"
            % (blob_path, args["--blobdir"], blob_path)
        )

        cmds.append("blobtools add --key static_plots=true %s" % blob_path)

        for cmd in cmds:
            logger.info(cmd)
            with subprocess.Popen(
                shlex.split(cmd),
                stdout=subprocess.PIPE,
                preexec_fn=os.setsid,
                encoding="utf-8",
            ) as process:
                try:
                    process.communicate(timeout=1800)[0]
                except subprocess.TimeoutExpired:
                    os.killpg(
                        process.pid, signal.SIGINT
                    )  # send signal to the process group
                    process.communicate()[0]

        for filename in os.listdir(blob_path):
            p = Path("%s/%s" % (blob_path, filename))
            parts = filename.split(".")
            if filename.startswith(args["--blobdir"]):
                if (
                    filename.endswith("png")
                    or filename.endswith("svg")
                    or filename.endswith("json")
                ):
                    if parts[1].isdigit():
                        parts = parts[2:]
                    else:
                        parts = parts[1:]
                    new_p = Path(
                        "%s/%s"
                        % (
                            p.parent.as_posix(),
                            filename.replace("%s." % args["--blobdir"], ""),
                        )
                    )
                    p.rename(new_p)
    except Exception as err:
        logger.error(err)
        for pngpath in glob.iglob(
            os.path.join(blob_path, "%s.*.png" % args["--blobdir"])
        ):
            os.remove(pngpath)
        for svgpath in glob.iglob(
            os.path.join(blob_path, "%s.*.svg" % args["--blobdir"])
        ):
            os.remove(svgpath)
        for jsonpath in glob.iglob(
            os.path.join(blob_path, "%s.*.json" % args["--blobdir"])
        ):
            os.remove(jsonpath)
        exit(1)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
Unchunk BLAST results.

Update coordinates and remove seq name suffix from chunked blast results.

Usage: blobtoolkit-pipeline unchunk-blast [--count INT] --in TSV --out TSV

Options:
    --in TSV     input filename.
    --out TSV    output filename.
    --count INT  number of results to keep per chunk. [Default: 10]
"""

import logging
import re
import sys
from collections import defaultdict

from docopt import DocoptExit
from docopt import docopt

logger_config = {
    "level": logging.INFO,
    "format": "%(asctime)s [%(levelname)s] line %(lineno)d %(message)s",
    "filemode": "w",
}
logging.basicConfig(**logger_config)
logger = logging.getLogger()


# def parse_args():
#     """Parse snakemake args if available."""
#     args = {}
#     try:
#         args["--in"] = snakemake.input[0]
#         args["--count"] = str(snakemake.params.max_target_seqs)
#         args["--out"] = snakemake.output[0]
#         for key, value in args.items:
#             sys.argv.append(key)
#             sys.argv.append(value)
#     except NameError as err:
#         pass


def main(rename=None):
    """Entry point."""
    docs = __doc__
    if rename is not None:
        docs = docs.replace("blobtoolkit-pipeline", rename)
    try:
        args = docopt(docs)
    except DocoptExit as e:
        raise DocoptExit from e
    try:
        lines = defaultdict(dict)
        chunk_counts = defaultdict(int)
        with open(args["--in"], "r") as fh:
            for line in fh.readlines():
                if "\n" not in line:
                    line += "\n"
                fields = line.split("\t")
                if fields[0]:
                    name, start = re.split("_-_", fields[0])
                    fields[0] = name
                    fields[3] = name
                    fields[9] = str(int(fields[9]) + int(start))
                    fields[10] = str(int(fields[10]) + int(start))
                    if start not in lines[name]:
                        lines[name][start] = []
                        chunk_counts[name] += 1
                    lines[name][start].append("\t".join(fields))
        with open(args["--out"], "w") as ofh:
            for name in chunk_counts.keys():
                length = len(lines[name])
                n = int(args["--count"])
                for start in lines[name].keys():
                    for i in range(n):
                        if i < len(lines[name][start]):
                            ofh.write("%s" % (lines[name][start][i]))
    except Exception as err:
        logger.error(err)
        exit(1)



#!/usr/bin/env python3
"""
Run BlobToolKit Pipeline.

Usage: blobtoolkit-pipeline run [--dry-run] [--tool STRING] [--unlock]
                        [--workdir DIR] [--threads INT] --config YAML

Options:
    --config YAML  YAML format configuration filename.
    --dry-run      Dry run flag.
    --threads INT  Number of threads to use. [Default: 32]
    --tool STRING  Pipeline tool to run. [Default: blobtoolkit]
    --unlock       Flag to unlock working directory.
    --workdir DIR  Full Path of working directory.
                   Default is the directory containing config file.
"""

import logging
import os
import shlex
import shutil
import subprocess

from docopt import DocoptExit
from docopt import docopt

logger_config = {
    "level": logging.INFO,
    "format": "%(asctime)s [%(levelname)s] line %(lineno)d %(message)s",
    "filemode": "w",
}
logging.basicConfig(**logger_config)
logger = logging.getLogger()


def run_command(cmd):
    """Run a command and capture CTRL-C."""
    script_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
    p = subprocess.Popen(shlex.split(cmd), stdin=subprocess.PIPE, cwd=script_dir)
    exit_code = p.wait()
    return exit_code


def unlock_working_directory(workdir):
    """Unlock Snakemake working directory after a failed run."""
    script_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
    snakefile = os.path.join(script_dir, "blobtoolkit-snakefiles", "blobtoolkit.smk")
    cmd = """
    snakemake -p \
          -j 1 \
          --directory %s/blobtoolkit \
          --configfile %s/config.yaml \
          --unlock \
          -s %s
    """ % (
        workdir,
        workdir,
        snakefile,
    )
    exit_code = run_command(cmd)
    return exit_code


def run_pipeline(workdir, args):
    """Run Snakemake pipeline."""
    script_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
    snakefile = os.path.join(
        script_dir, "blobtoolkit-snakefiles", f'{args["--tool"]}.smk'
    )
    dry_run = "-n" if "--dry-run" in args and args["--dry-run"] is True else ""
    cmd = """
    snakemake -p %s \
          -j %s \
          --directory %s/blobtoolkit \
          --configfile %s/config.yaml \
          -s %s
    """ % (
        dry_run,
        args["--threads"],
        workdir,
        workdir,
        snakefile,
    )
    exit_code = run_command(cmd)
    return exit_code


def run_snakemake_pipeline(args):
    config = os.path.abspath(args["--config"])
    workdir = (
        os.path.abspath(args["--workdir"])
        if args["--workdir"] is not None
        else os.path.abspath(os.path.dirname(config))
    )
    if config != f"{workdir}/config.yaml":
        shutil.copy2(config, f"{workdir}/config.yaml")
    if args["--unlock"]:
        exit_code = unlock_working_directory(workdir)
        exit(exit_code)
    exit_code = run_pipeline(workdir, args)
    exit(exit_code)


def main(rename=None):
    """Entry point."""
    docs = __doc__
    if rename is not None:
        docs = docs.replace("blobtoolkit-pipeline", rename)
    try:
        args = docopt(docs)
    except DocoptExit as e:
        raise DocoptExit from e
    try:
        run_snakemake_pipeline(args)
    except Exception as err:
        logger.error(err)
        exit(1)



#!/usr/bin/env python3
"""
Extract BUSCO gene sequences and set headers.

Usage: blobtoolkit-pipeline extract-busco-genes --busco PATH... --out FASTA

Options:
    --busco PATH         BUSCO full summary tsv file or busco_sequences directory.
    --out FASTA          output FASTA filename.
"""

import codecs
import glob
import logging
import re
import tarfile
from pathlib import Path

from docopt import DocoptExit
from docopt import docopt

logger_config = {
    "level": logging.INFO,
    "format": "%(asctime)s [%(levelname)s] line %(lineno)d %(message)s",
    "filemode": "w",
}
logging.basicConfig(**logger_config)
logger = logging.getLogger()


# def parse_args():
#     """Parse snakemake args if available."""
#     args = {}
#     try:
#         args["--busco"] = snakemake.input.busco
#         args["--out"] = snakemake.output.fasta
#     except NameError as err:
#         pass
#     for key, value in args.items:
#         sys.argv.append(key)
#         sys.argv.append(value)


def parse_busco_file(fh, ofh, status, busco_id):
    """Parse a busco sequences file."""
    header_pattern = re.compile(r">\S+:\d+-\d+")
    for line in fh:
        if line.startswith(">"):
            header = line.strip()
            title = ""
            if header_pattern.match(header):
                title = f"{header}={busco_id}={status}"
            else:
                parts = header.split(" # ")
                title = f'{re.sub("_[^_]+$", "", parts[0])}:{parts[1]}-{parts[2]}={busco_id}={status}'
            ofh.write("%s\n" % title)
        else:
            ofh.write(line)


def main(rename=None):
    """Entry point."""
    docs = __doc__
    if rename is not None:
        docs = docs.replace("blobtoolkit-pipeline", rename)
    try:
        args = docopt(docs)
    except DocoptExit as e:
        raise DocoptExit from e
    file_pattern = re.compile(r"busco_sequences\/(\w+?)_\w+\/(\d+at\d+)\.faa")
    try:
        with open(args["--out"], "w") as ofh:
            for busco_file in args["--busco"]:
                if busco_file.endswith("busco_sequences"):
                    # plain directory
                    parse_busco_directory(file_pattern, ofh, busco_file)
                else:
                    # tarred archive
                    utf8reader = codecs.getreader("utf-8")
                    busco_dir = Path(busco_file).absolute().parent
                    busco_seqs = f"{busco_dir}/busco_sequences.tar.gz"
                    if not Path(busco_seqs).is_file():
                        continue
                    parse_tarred_busco_directory(
                        file_pattern, ofh, utf8reader, busco_seqs
                    )

    except Exception as err:
        logger.error(err)
        exit(1)


def parse_tarred_busco_directory(file_pattern, ofh, utf8reader, busco_seqs):
    """Parse a tarred busco directory."""
    tar = tarfile.open(busco_seqs)
    for tarinfo in tar.getmembers():
        if tarinfo.name.endswith(".faa"):
            match = file_pattern.match(tarinfo.name)
            status, busco_id = match.groups()
            with utf8reader(
                tar.extractfile(
                    tarinfo,
                )
            ) as fh:
                parse_busco_file(fh, ofh, status, busco_id)


def parse_busco_directory(file_pattern, ofh, busco_file):
    """Parse a plain busco_directory."""
    files = glob.glob(f"{busco_file}/*/*.faa")
    for file in files:
        match = file_pattern.search(file)
        status, busco_id = match.groups()
        with open(file, "r") as fh:
            parse_busco_file(fh, ofh, status, busco_id)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

"""
Window stats.

Usage:
    blobtoolkit-pipeline window-stats --in TSV [--window FLOAT...]
        [--min-window-length INT] [--min-window-count INT]
        [--headers STRING] --out TSV

Options:
    --in TSV                 chunked summary stats tsv file.
    --min-window-length INT  minimum length of a window. [Default: 100000]
    --min-window-count INT   minimum number of windows. [Default: 1]
    --window FLOAT           window size or proportion. [Default: 1]
    --headers STRING         list of headers to use. [Default: all]
    --out TSV                output TSV filename or suffix.
"""


import contextlib
import logging
import math
import os
import re
import statistics
from collections import defaultdict
from pathlib import Path

from docopt import DocoptExit
from docopt import docopt
from tolkein import tofile

logger_config = {
    "level": logging.INFO,
    "format": "%(asctime)s [%(levelname)s] line %(lineno)d %(message)s",
    "filemode": "w",
}
logging.basicConfig(**logger_config)
logger = logging.getLogger()


def set_header_keys(header, args):
    """Set header keys."""
    header_keys = {}
    if args["--headers"] == "all":
        header_keys = {key: key for key in header.keys()}
    else:
        heads = args["--headers"].split(",")
        for head in heads:
            try:
                orig, trans = head.split("=")
                header_keys[orig] = trans
            except ValueError:
                header_keys[head] = head
    return header_keys


def parse_chunked_values(filename, args):
    """Parse chunked values into dict."""
    interval = 0
    header = None
    values = defaultdict(lambda: defaultdict(list))
    lengths = defaultdict(int)
    with tofile.open_file_handle(filename) as fh:
        for line in fh.readlines():
            row = line.rstrip().split("\t")
            if header is None:
                header = {key: idx + 3 for idx, key in enumerate(row[3:])}
                header_keys = set_header_keys(header, args)
                continue
            seqid = row[0]
            chunk_length = int(row[2]) - int(row[1])
            if chunk_length > interval:
                interval = chunk_length
            lengths[seqid] += chunk_length
            for key, idx in header.items():
                if key in header_keys:
                    with contextlib.suppress(ValueError):
                        values[seqid][header_keys[key]].append(float(row[idx]))
    return lengths, values, interval


def round_to_interval(value, interval):
    """Round a number to a given interval."""
    return round(
        round(value / interval + 0.5) * interval, -int(math.floor(math.log10(interval)))
    )


def get_window_size(length, interval, window, min_length, min_count):
    """Get size of window based on options."""
    if window == 1:
        return length
    if window < 1:
        window_length = round_to_interval(length * window, interval)
        return window_length if window_length > min_length else None
    if length / window >= min_count:
        return round_to_interval(window, interval)
    return None


def calculate_mean(arr, log):
    """Calculate mean and sd of arr values."""
    n = len(arr)
    if log:
        logged_arr = [math.log10(value) for value in arr if value > 0]
        if not logged_arr:
            return 0, 0, n
        mean = math.pow(10, statistics.mean(logged_arr))
        sd = math.pow(10, statistics.stdev(logged_arr)) if len(logged_arr) > 1 else 0
        return mean, sd, n
    mean = statistics.mean(arr)
    sd = statistics.stdev(arr) if n > 1 else 0
    return mean, sd, n


def combine_chunks(window, interval, values, seqid, length, window_size, key, arr):
    start_i = 0
    start_pos = 0
    while start_pos < length:
        end_pos = min(start_pos + window_size, length)
        mid_pos = start_pos + (end_pos - start_pos) / 2
        end_i = round(end_pos / interval + 0.5) - 1
        proportion = "1" if window == 1 else "%.3f" % (mid_pos / length)
        if start_pos not in values[seqid]:
            values[seqid][start_pos] = {
                "end": str(end_pos),
                "proportion": proportion,
            }
        if key.endswith("count"):
            values[seqid][start_pos][key] = "%d" % sum(arr[start_i:end_i])
        else:
            try:
                mean, sd, n = calculate_mean(
                    arr[start_i : end_i + 1], key.endswith("_cov")
                )
                values[seqid][start_pos][key] = "%.3f" % mean
                values[seqid][start_pos][f"{key}_sd"] = "%.3f" % sd
                values[seqid][start_pos][f"{key}_n"] = "%d" % n
            except statistics.StatisticsError:
                values[seqid][start_pos][key] = "None"
                values[seqid][start_pos][f"{key}_sd"] = "None"
                values[seqid][start_pos][f"{key}_n"] = "None"
        if end_pos == start_pos:
            break
        start_pos = end_pos
        start_i = end_i + 1


def calculate_window_stats(lengths, chunks, window, interval, args):
    """Calculate mean and sd in windows across each sequence."""
    values = defaultdict(dict)
    for seqid, length in lengths.items():
        window_size = get_window_size(
            length,
            interval,
            window,
            int(args["--min-window-length"]),
            int(args["--min-window-count"]),
        )
        if window_size is not None:
            for key, arr in chunks[seqid].items():
                combine_chunks(
                    window, interval, values, seqid, length, window_size, key, arr
                )
    return values


def process_files(args):
    lengths, chunks, interval = parse_chunked_values(args["--in"], args)
    outfile = args["--out"]
    if outfile.startswith("."):
        outfile = f'{args["--in"]}{args["--out"]}'
    outfile = Path(outfile)
    suffix = outfile.suffix
    filename = outfile.stem
    if filepath := outfile.parent:
        filename = f"{filepath}/{filename}"
    for window in args["--window"]:
        window = float(window)
        values = calculate_window_stats(lengths, chunks, window, interval, args)
        rows = []
        header = None
        for seqid, obj in values.items():
            for start_pos, entry in obj.items():
                if header is None:
                    header = ["sequence", "start", "end"]
                    for key in entry.keys():
                        if key not in header:
                            header.append(key)
                    rows.append("\t".join(header) + "\n")
                row = [seqid, str(start_pos)]
                row.extend(entry[key] for key in header[2:])
                rows.append("\t".join(row) + "\n")
        if rows:
            filetag = ""
            if window != 1:
                filetag = ".%s" % re.sub(r"\.0$", "", str(window))
            if not os.path.exists(os.path.dirname(filename)):
                os.makedirs(os.path.dirname(filename))
            with open(f"{filename}{filetag}{suffix}", "w") as fh:
                fh.writelines(rows)


def main(rename=None):
    """Entry point."""
    docs = __doc__
    if rename is not None:
        docs = docs.replace("blobtoolkit-pipeline", rename)
    try:
        args = docopt(docs)
    except DocoptExit as e:
        raise DocoptExit from e
    try:
        process_files(args)
    except Exception as err:
        logger.error(err)
        raise err


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
Split long sequences into chunks.

Usage: blobtoolkit-pipeline chunk-fasta --in FASTA [--chunk INT] [--overlap INT] [--max-chunks INT]
                [--busco TSV] [--min-length INT] [--out FASTA] [--bed BEDFILE]

Options:
    --in FASTA           input FASTA file.
    --busco TSV          BUSCO full summary tsv file.
    --chunk INT          sequences greater than CHUNK bp will be split. [Default: 100000]
    --overlap INT        length of overlap when splitting sequences. [Default: 0]
    --max-chunks INT     maximum number of chunks to split a sequence into. [Default: 10]
    --min-length INT     minimum sequence length. [Default: 1000]
    --out FASTA          output FASTA filename or suffix. [Default: .chunked]
    --bed BEDFILE        output BED/BED-like TSV filename or suffix. [Default: .bed]
"""

import logging
import math
import re
import shlex
import sys
from collections import Counter
from collections import defaultdict
from itertools import groupby
from subprocess import PIPE
from subprocess import Popen

from docopt import DocoptExit
from docopt import docopt
from tolkein import tofile

logger_config = {
    "level": logging.INFO,
    "format": "%(asctime)s [%(levelname)s] line %(lineno)d %(message)s",
    "filemode": "w",
}
logging.basicConfig(**logger_config)
logger = logging.getLogger()


def chunk_size(value):
    """Calculate nice value for chunk size."""
    return round(value / 1000 + 0.5) * 1000


def chunk_fasta(
    fastafile, *, chunk=math.inf, overlap=0, max_chunks=math.inf, min_length=0
):
    """Read FASTA file one sequence at a time and split long sequences into chunks."""
    cmd = "cat %s" % fastafile
    if fastafile.endswith(".gz"):
        cmd = "zcat %s" % fastafile
    title = ""
    seq = ""
    segment = chunk + overlap
    with Popen(shlex.split(cmd), encoding="utf-8", stdout=PIPE, bufsize=4096) as proc:
        faiter = (x[1] for x in groupby(proc.stdout, lambda line: line[0] == ">"))
        for header in faiter:
            title = header.__next__()[1:].strip().split()[0]
            seq = "".join(map(lambda s: s.strip(), faiter.__next__()))
            seq_length = len(seq)
            my_chunk = chunk
            if seq_length > segment:
                n = (seq_length + chunk) // chunk
                if n > max_chunks:
                    my_chunk = chunk_size(seq_length / max_chunks)
                    n = max_chunks
                for i in range(0, seq_length, my_chunk):
                    subseq = seq[i : i + my_chunk + overlap]
                    yield {
                        "title": title,
                        "seq": subseq,
                        "chunks": n,
                        "start": i,
                        "end": i + my_chunk + overlap,
                        "length": my_chunk + overlap,
                    }
            elif seq_length > min_length:
                yield {
                    "title": title,
                    "seq": seq,
                    "chunks": 1,
                    "start": 0,
                    "end": seq_length,
                    "length": seq_length,
                }


def check_for_unmasked_bases(seq, min_unmasked=20):
    """Check sequence has runs of unmasked bases."""
    return bool(re.search(r"[ACGT]{" + str(min_unmasked) + "}", seq))


def check_for_masked_bases(seq, max_masked=20):
    """Check sequence has runs of masked bases."""
    return bool(re.search(r"[acgtnN]{" + str(max_masked) + "}", seq))


def parse_busco_full_summary(busco_file, chunk=100000):
    """Parse a busco full summary file."""
    logger.info("Parsing BUSCO full summary file")
    locations = defaultdict(list)
    with tofile.open_file_handle(busco_file) as fh:
        for line in fh:
            if line.startswith("#"):
                continue
            parts = line.split("\t")
            if parts[1] in ("Complete", "Duplicated"):
                if ":" in parts[2] and parts[5] == "-":
                    start = int(parts[4])
                    end = int(parts[3])
                else:
                    start = int(parts[3])
                    end = int(parts[4])
                locations[parts[2].split(":")[0]].append((start, end))
    windows = {}
    for title, tuples in locations.items():
        tuples.sort(key=lambda tup: tup[0])
        windows[title] = []
        start_index = 0
        for location in tuples:
            windows[title].append([location[0], location[0] + chunk, 0])
            for window in windows[title][start_index:]:
                if location[1] < window[1]:
                    window[2] += 1
                else:
                    start_index += 1
        windows[title].sort(key=lambda window: window[2], reverse=True)
    return windows


def chunk_by_busco(seq, seqs, busco_windows, args):
    """Use BUSCO gene density to define chunks."""
    subseq_found = False
    has_busco = False
    if seq["title"] in busco_windows:
        for window in busco_windows[seq["title"]]:
            if seq["start"] <= window[0] and seq["end"] >= window[1]:
                has_busco = True
                subseq = seq["seq"][window[0] - seq["start"] : window[1] - seq["start"]]
                # check there are runs of unmasked bases
                if check_for_unmasked_bases(subseq, 1000):
                    seq["seq"] = subseq
                    seq["start"] = window[0]
                    subseq_found = True
                    break
    if not has_busco:
        # extract subseq from middle of chunk
        chunk = int(args["--chunk"])
        offset = int((seq["length"] - chunk) / 2)
        if offset > 0:
            midseq = seq["seq"][offset:-offset]
            # check there are runs of unmasked bases
            if check_for_unmasked_bases(midseq, 1000):
                seq["seq"] = midseq
                seq["start"] += offset
                seq["end"] -= offset
                subseq_found = True
            else:
                # walk along sequence to find a chunk with fewer masked bases
                while offset > chunk:
                    offset -= chunk
                    lseq = seq["seq"][offset : offset + chunk]
                    if check_for_unmasked_bases(lseq, 1000):
                        seq["seq"] = lseq
                        seq["start"] += offset
                        seq["end"] = seq["start"] + chunk
                        subseq_found = True
                        break
                    rseq = seq["seq"][-offset - chunk : -offset]
                    if check_for_unmasked_bases(rseq, 1000):
                        seq["seq"] = rseq
                        seq["start"] = seq["end"] - offset - chunk
                        seq["end"] -= offset
                        subseq_found = True
                        break
    if subseq_found:
        seqs.append((seq))


def seq_stats(seq):
    """Calculate basic sequence statistics."""
    counter = Counter(seq)
    gc = 0
    at = 0
    n = 0
    masked = 0
    for base in ["g", "c", "G", "C"]:
        if base in counter:
            gc += counter[base]
    for base in ["a", "t", "A", "T"]:
        if base in counter:
            at += counter[base]
    for base in ["n", "N"]:
        if base in counter:
            n += counter[base]
    for base in ["a", "c", "g", "t", "n"]:
        if base in counter:
            masked += counter[base]
    atgc = at + gc
    if atgc > 0:
        gc = gc / atgc
    else:
        gc = 0
    return {"gc": gc, "n": n / len(seq), "ncount": n, "masked": masked / len(seq)}


def write_bedfiles(bed_data, args):
    """Write bedfiles."""
    if "--bed" not in args or not args["--bed"]:
        return
    locations = []
    rows = []
    header = ""
    for title, arr in bed_data.items():
        for obj in arr:
            if not header:
                header = "sequence\tstart\tend\t%s\n" % "\t".join(obj["stats"].keys())
                rows.append(header)
            row = "%s\t%d\t%d" % (title, obj["start"], obj["end"])
            locations.append("%s\n" % row)
            for key, value in obj["stats"].items():
                if key != "ncount":
                    row += "\t%.4f" % value
                else:
                    row += "\t%d" % value
            rows.append("%s\n" % row)
    bedfile = args["--bed"]
    if bedfile.startswith("."):
        bedfile = "%s%s" % (args["--in"], bedfile)
    bedfile = re.sub(r"\.bed$", "", bedfile)
    bedfile = re.sub(r"\.tsv$", "", bedfile)
    with open("%s.mask.bed" % bedfile, "w") as ofh:
        ofh.writelines(locations)
    with open("%s.tsv" % bedfile, "w") as ofh:
        ofh.writelines(rows)


# def parse_args():
#     """Parse snakemake args if available."""
#     args = {}
#     try:
#         args["--in"] = snakemake.input.fasta
#         args["--chunk"] = str(snakemake.params.chunk)
#         args["--overlap"] = str(snakemake.params.overlap)
#         args["--max-chunks"] = str(snakemake.params.max_chunks)
#         args["--min-length"] = str(snakemake.params.min_length)
#         try:
#             args["--busco"] = snakemake.input.busco
#         except AttributeError:
#             args["--busco"] = "None"
#         try:
#             args["--out"] = snakemake.output.fasta
#         except AttributeError:
#             args["--out"] = "None"
#         try:
#             args["--bed"] = snakemake.params.bed
#         except AttributeError:
#             args["--bed"] = "None"
#         for key, value in args.items():
#             sys.argv.append(key)
#             sys.argv.append(value)
#     except NameError as err:
#         pass


def main(rename=None):
    """Entry point."""
    docs = __doc__
    if rename is not None:
        docs = docs.replace("blobtoolkit-pipeline", rename)
    try:
        args = docopt(docs)
    except DocoptExit as e:
        raise DocoptExit from e
    try:
        make_chunks(args)
    except Exception as err:
        logger.error(err)
        exit(1)


def make_chunks(args):
    busco_windows = (
        parse_busco_full_summary(args["--busco"])
        if (
            "--busco" in args
            and args["--busco"] is not None
            and args["--busco"] != "None"
        )
        else {}
    )
    logger.info(f'Splitting {args["--in"]} into chunks')
    bed_data = defaultdict(list)
    seqs = []
    for seq in chunk_fasta(
        args["--in"],
        chunk=int(args["--chunk"]),
        overlap=int(args["--overlap"]),
        max_chunks=int(args["--max-chunks"]),
    ):
        if (
            busco_windows
            and seq["chunks"] == int(args["--max-chunks"])
            and seq["length"] > int(args["--chunk"]) + int(args["--overlap"])
        ):
            chunk_by_busco(seq, seqs, busco_windows, args)
        else:
            if "--bed" in args and args["--bed"] and args["--bed"] != "None":
                stats = seq_stats(seq["seq"])
                bed_data[seq["title"]].append(
                    {
                        "start": seq["start"],
                        "end": seq["start"] + len(seq["seq"]),
                        "stats": stats,
                    }
                )
            if check_for_unmasked_bases(seq["seq"]):
                seqs.append((seq))
    if bed_data:
        write_bedfiles(bed_data, args)
    if args["--out"] is not None and args["--out"] != "None":
        chunked = ""
        for seq in seqs:
            if seq["length"] >= int(args["--min-length"]):
                chunked += ">%s_-_%d\n" % (seq["title"], seq["start"])
                chunked += "%s\n" % seq["seq"]
        outfile = args["--out"]
        if outfile.startswith("."):
            outfile = f'{args["--in"]}{args["--out"]}'
        with open(outfile, "w") as ofh:
            ofh.writelines(chunked)


if __name__ == "__main__":
    main()
